from typing import Any, Dict

from flask import Flask
from flask_marshmallow import Marshmallow
from marshmallow import fields, post_load

from .model import AccessToken

__all__ = ["new_token_schema", "token_schema", "ma", "setup_schemas"]

ma = Marshmallow()


def setup_schemas(app: Flask):
    return ma.init_app(app)


class AccessTokenSchema(ma.Schema):
    class Meta:
        ordered = True
    id = fields.UUID(required=False)
    user_id = fields.UUID()
    name = fields.String()
    access_token = fields.String()
    refresh_token = fields.String()
    url = ma.AbsoluteURLFor('token.get', token_id='<id>')
    links = ma.Hyperlinks({
        'self': ma.URLFor('token.get', token_id='<id>'),
    })

    @post_load
    def make_access_token(self, data: Dict[str, Any]):
        return AccessToken(**data)


class NewAccessTokenSchema(ma.Schema):
    name = fields.String(required=True)
    user_id = fields.String(required=True)
    scopes = fields.String(many=True, default=None)


token_schema = AccessTokenSchema()
new_token_schema = NewAccessTokenSchema()



from typing import Any, Dict

from chaosplt_grpc import remote_channel
from chaosplt_grpc.registration.client import create_registration, \
    get_by_id

from ..model import User

__all__ = ["AccountService"]


class AccountService:
    def __init__(self, config: Dict[str, Any]):
        self.account_addr = config["grpc"]["account"]["addr"]
        self.user = UserService(self.account_addr)
        self.registration = RegistrationService(self.account_addr)

    def release(self):
        pass


class UserService:
    def __init__(self, addr: str):
        self.addr = addr


class RegistrationService:
    def __init__(self, addr: str):
        # gRPC address of the remote service
        self.addr = addr

    def create(self, username: str, name: str, email: str) -> User:
        """
        Register a new user.
        """
        with remote_channel(self.addr) as channel:
            registration = create_registration(channel, username, name, email)
            return User(
                registration.id, registration.username, registration.name,
                registration.email, is_active=registration.is_active,
                is_authenticated=registration.is_authenticated,
                is_anonymous=registration.is_anonymous)

    def get(self, registration_id: str) -> User:
        """
        Retrieve an user by its registration identifier.
        """
        with remote_channel(self.addr) as channel:
            registration = get_by_id(channel, registration_id)
            return User(
                registration.id, registration.username, registration.name,
                registration.email, is_active=registration.is_active,
                is_authenticated=registration.is_authenticated,
                is_anonymous=registration.is_anonymous)



from uuid import UUID

import attr

__all__ = ["AccessToken", "OAuthToken", "User"]


@attr.s
class AccessToken:
    name: str = attr.ib()
    id: UUID = attr.ib()
    user_id: UUID = attr.ib()
    access_token: str = attr.ib()
    refresh_token: str = attr.ib()


@attr.s
class OAuthToken:
    id: UUID = attr.ib()
    user_id: UUID = attr.ib()
    provider: str = attr.ib()
    provider_id = attr.ib()
    token = attr.ib()


@attr.s
class User:
    id: UUID = attr.ib()
    username: str = attr.ib()
    name: str = attr.ib()
    email: str = attr.ib()
    is_authenticated: bool = attr.ib(default=False)
    is_active: bool = attr.ib(default=False)
    is_anonymous: bool = attr.ib(default=True)

    def get_id(self) -> UUID:
        return self.id



from abc import ABC, abstractmethod
from typing import Dict, NoReturn, Union
from uuid import UUID

import attr

from chaosplt_auth.model import AccessToken, OAuthToken

__all__ = ["BaseAuthStorage", "BaseAccessTokenService",
           "BaseOAuthTokenService"]


class BaseAccessTokenService(ABC):
    @abstractmethod
    def get(self, user_id: Union[UUID, str],
            token_id: Union[UUID, str]) -> AccessToken:
        """
        Fetch a token for the given user
        """
        raise NotImplementedError()

    @abstractmethod
    def get_by_user(self, user_id: Union[UUID, str]) -> Dict[str, AccessToken]:
        """
        Fetch all tokens for the given user.

        The returned dictionary uses the tokens' name as keys.
        """
        raise NotImplementedError()

    @abstractmethod
    def create(self, name: str, user_id: Union[UUID, str]) -> AccessToken:
        """
        Create a new token with the given name for the user
        """
        raise NotImplementedError()

    @abstractmethod
    def delete(self, user_id: Union[UUID, str],
               token_id: Union[UUID, str]) -> NoReturn:
        """
        Delete a token from the given user
        """
        raise NotImplementedError()


class BaseOAuthTokenService(ABC):
    @abstractmethod
    def get(self, oauth_id: Union[UUID, str]) -> OAuthToken:
        """
        Fetch a token for the given user
        """
        return self.load(oauth_id)

    @abstractmethod
    def create(self, user_id: str, provider: str, provider_id: str,
               token: str) -> OAuthToken:
        """
        Create a new token
        """
        raise NotImplementedError()

    @abstractmethod
    def delete(self, oauth_id: str) -> NoReturn:
        """
        Delete the given token
        """
        raise NotImplementedError()

    @abstractmethod
    def delete_by_user(self, user_id: Union[UUID, str]) -> NoReturn:
        """
        Delete the OAuth token of the given user.
        """
        raise NotImplementedError()

    @abstractmethod
    def set_for_user(self, user_id: Union[UUID, str], token: str) -> NoReturn:
        """
        Set the OAuth token of the given user.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_by_user(self, user_id: Union[UUID, str]) -> OAuthToken:
        """
        Get the OAuth token of the given user.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_by_provider(self, provider: str, provider_id: str) -> OAuthToken:
        """
        Get the OAuth token for the given provider/provider-id pair which
        should be uniq.
        """
        raise NotImplementedError()


@attr.s
class BaseAuthStorage:
    access_token: BaseAccessTokenService = attr.ib()
    oauth_token: BaseOAuthTokenService = attr.ib()

    def release(self) -> NoReturn:
        raise NotImplementedError()



# -*- coding: utf-8 -*-
from typing import Dict, NoReturn, Union
import uuid
from uuid import UUID

from chaosplt_relational_storage.db import Base
from flask_dance.consumer.backend.sqla import OAuthConsumerMixin
from sqlalchemy import Boolean, Column, DateTime, String, \
    UniqueConstraint, func
from sqlalchemy.orm.session import Session
from sqlalchemy_utils import UUIDType

__all__ = ["AccessToken", "OAuthToken"]


class AccessToken(Base):  # type: ignore
    __tablename__ = 'access_token'
    __table_args__ = (
        UniqueConstraint(
            'name', 'user_id'
        ),
    )

    id = Column(
        UUIDType(binary=False), primary_key=True, default=uuid.uuid4)
    name = Column(String(128), nullable=False)
    user_id = Column(UUIDType(binary=False), nullable=False)
    revoked = Column(Boolean(name='revoked'), nullable=False, default=False)
    issued_on = Column(DateTime(), nullable=False, server_default=func.now())
    last_used_on = Column(DateTime(), nullable=True)
    access_token = Column(String(255), index=True, nullable=False)
    refresh_token = Column(String(255), index=True, nullable=False)

    @staticmethod
    def save(name: str, user_id: Union[UUID, str], access_token: str,
             refresh_token: str, session: Session) -> 'AccessToken':
        token = AccessToken(
            name=name,
            user_id=user_id,
            access_token=access_token,
            refresh_token=refresh_token)

        session.add(token)
        return token

    @staticmethod
    def delete(user_id: Union[UUID, str],
               token_id: Union[UUID, str], session: Session) -> NoReturn:
        token = session.query(AccessToken).\
            filter_by(user_id=user_id).\
            filter_by(id=token_id).first()

        if token:
            session.delete(token)

    @staticmethod
    def load_by_user(user_id: Union[UUID, str],
                     session: Session) -> Dict[str, 'AccessToken']:
        return session.query(AccessToken).\
            filter_by(user_id=user_id).all()

    @staticmethod
    def load(user_id: Union[UUID, str], token_id: Union[UUID, str],
             session: Session) -> 'AccessToken':
        return session.query(AccessToken).\
            filter_by(user_id=user_id).\
            filter_by(id=token_id).\
            filter_by(revoked=False).first()


class OAuthToken(Base, OAuthConsumerMixin):
    __tablename__ = 'oauth_token'
    __table_args__ = (
        UniqueConstraint(
            'provider', 'provider_id'
        ),
    )

    provider_id = Column(String(256), unique=True, nullable=False)
    user_id = Column(UUIDType(binary=False), nullable=False)

    @staticmethod
    def load(user_id: Union[UUID, str], oauth_id: Union[UUID, str],
             session: Session) -> 'OAuthToken':
        return session.query(OAuthToken).\
            filter_by(user_id=user_id).\
            filter_by(id=oauth_id).first()

    @staticmethod
    def load_by_user(user_id: Union[UUID, str],
                     session: Session) -> 'OAuthToken':
        return session.query(OAuthToken).filter_by(
            user_id=user_id).first()

    @staticmethod
    def load_by_provider(provider: str, provider_id: str,
                         session: Session) -> 'OAuthToken':
        return session.query(OAuthToken).\
            filter_by(provider=provider).\
            filter_by(provider_id=provider_id).first()

    @staticmethod
    def delete(user_id: Union[UUID, str], oauth_id: Union[UUID, str],
               session: Session) -> NoReturn:
        return session.query(OAuthToken).\
            filter_by(user_id=user_id).\
            filter_by(id=oauth_id).delete()

    @staticmethod
    def delete_by_user(user_id: Union[UUID, str],
                       session: Session) -> NoReturn:
        return session.query(OAuthToken).filter_by(
            user_id=user_id).delete()

    @staticmethod
    def save(user_id: Union[UUID, str], provider: str, provider_id: str,
             token: str, session: Session) -> 'OAuthToken':
        oauth = OAuthToken(
            user_id=user_id,
            token=token,
            provider=provider,
            provider_id=provider_id
        )

        session.add(oauth)
        return oauth

    @staticmethod
    def update_by_user(user_id: Union[UUID, str], token: str,
                       session: Session) -> NoReturn:
        oauth = session.query(OAuthToken).filter_by(
            user_id=user_id).first()
        if oauth:
            oauth.token = token



# -*- coding: utf-8 -*-
import logging
from logging import StreamHandler
import os
from typing import Any, Dict

from flask import Blueprint, Flask, request, after_this_request, Response
from flask_caching import Cache

from chaosplt_auth.auth import setup_login
from chaosplt_auth.service import Services
from chaosplt_auth.storage import AuthStorage

from .backend import setup_oauth_provider
from .oauth2 import view as oauth2_view, github_view

__all__ = ["create_app", "cleanup_app", "serve_app"]


def create_app(config: Dict[str, Any]) -> Flask:
    app = Flask(__name__)

    app.url_map.strict_slashes = False
    app.debug = True if os.getenv('CHAOSHUB_DEBUG') else False

    logger = logging.getLogger('flask.app')
    logger.propagate = False

    app.config["SECRET_KEY"] = os.getenv("SECRET_KEY")
    app.config["JWT_SECRET_KEY"] = os.getenv("JWT_SECRET_KEY")
    app.secret_key = app.config["SECRET_KEY"]

    app.config["GRPC_LISTEN_ADDR"] = os.getenv("GRPC_LISTEN_ADDR")

    app.config["CACHE_TYPE"] = os.getenv("CACHE_TYPE", "simple")
    if app.config["CACHE_TYPE"] == "redis":
        app.config["CACHE_REDIS_HOST"] = os.getenv("CACHE_REDIS_HOST")
        app.config["CACHE_REDIS_PORT"] = os.getenv("CACHE_REDIS_PORT", 6379)

    # OAUTH2
    app.config["GITHUB_OAUTH_CLIENT_ID"] = os.getenv("AUTH_GITHUB_CLIENT_ID")
    app.config["GITHUB_OAUTH_CLIENT_SECRET"] = os.getenv(
        "AUTH_GITHUB_CLIENT_SECRET")

    setup_login(app, from_session=True)

    return app


def cleanup_app(app: Flask):
    pass


def serve_app(app: Flask, cache: Cache, services: Services,
              storage: AuthStorage, config: Dict[str, Any],
              mount_point: str = '/oauth2', log_handler: StreamHandler = None):
    register_views(app, cache, services, storage)


###############################################################################
# Internals
###############################################################################
def register_views(app: Flask, cache: Cache, services, storage: AuthStorage):
    patch_request(oauth2_view, services, storage)
    patch_request(github_view, services, storage)

    app.register_blueprint(github_view, url_prefix="/auth/oauth2/via")
    app.register_blueprint(oauth2_view, url_prefix="/auth/oauth2")

    setup_oauth_provider(github_view, services, storage, cache)


def patch_request(bp: Blueprint, services, storage: AuthStorage):
    @bp.before_request
    def prepare_request():
        request.services = services
        request.storage = storage

        @after_this_request
        def clean_request(response: Response):
            request.services = None
            request.storage = None
            return response



# -*- coding: utf-8 -*-
import logging
from typing import Any, Dict, NoReturn, Union
from uuid import UUID

from flask import Blueprint
from flask_caching import Cache
from flask_dance.consumer import oauth_authorized, oauth_error
from flask_dance.consumer.backend import BaseBackend
from flask_login import login_user

from chaosplt_auth.model import OAuthToken, User
from chaosplt_auth.storage import AuthStorage

__all__ = ["setup_oauth_provider"]

logger = logging.getLogger("chaoshub")


class OAuthBackend(BaseBackend):
    def __init__(self, services, storage: AuthStorage, cache: Cache):
        self.services = services
        self.storage = storage
        self.cache = cache

    def make_cache_key(self, blueprint: Blueprint,
                       user_id: Union[UUID, str]) -> str:
        return "oauth_token|{name}|{user_id}".format(
            name=blueprint.name, user_id=user_id,
        )

    def get_oauth_token(self, user_id: Union[UUID, str]) -> OAuthToken:
        return self.storage.oauth_token.get_by_user(user_id)

    def get_oauth_token_by_provider(self, provider: str,
                                    provider_id: str) -> OAuthToken:
        return self.storage.oauth_token.get_by_provider(
            provider, provider_id)

    def delete_oauth_token(self, user_id: Union[UUID, str]) -> NoReturn:
        self.storage.oauth_token.delete_by_user(user_id)

    def set_oauth_token(self, user_id: Union[UUID, str],
                        token: str) -> OAuthToken:
        return self.storage.oauth_token.set_for_user(user_id, token)

    def create_oauth_token(self, user_id: Union[UUID, str], provider: str,
                           provider_id: str,
                           token: Dict[str, Any]) -> OAuthToken:
        return self.storage.oauth_token.create(
            user_id, provider, provider_id, token)

    def get_account(self, user_id: Union[UUID, str]) -> User:
        return self.services.account.registration.get(str(user_id))

    def create_account(self, username: str, name: str, email: str) -> User:
        return self.services.account.registration.create(
            username, name, email)

    def get(self, blueprint: Blueprint, user_id: Union[UUID, str]) -> str:
        cache_key = self.make_cache_key(
            blueprint=blueprint, user_id=user_id)
        token = self.cache.get(cache_key)
        if token:
            return token

        token = self.get_oauth_token(user_id)
        if token:
            value = token.token
            self.cache.set(cache_key, value)
            return value

    def set(self, blueprint: Blueprint, token: str,
            user_id: str) -> NoReturn:
        cache_key = self.make_cache_key(
            blueprint=blueprint, user_id=user_id)
        self.cache.delete(cache_key)
        self.delete_oauth_token(user_id)
        self.set_oauth_token(user_id, token)
        self.cache.set(cache_key, token)

    def delete(self, blueprint: Blueprint, user_id: str) -> NoReturn:
        cache_key = self.make_cache_key(
            blueprint=blueprint, user_id=user_id)
        self.cache.delete(cache_key)
        self.delete_oauth_token(user_id)


def setup_oauth_provider(blueprint: Blueprint, services, storage: AuthStorage,
                         cache: Cache):
    blueprint.backend = OAuthBackend(
        services=services, storage=storage, cache=cache)

    @oauth_authorized.connect_via(blueprint)
    def github_logged_in(blueprint: Blueprint, token: str):
        if not token:
            logger.error("Missing token")
            return False

        resp = blueprint.session.get("/user")
        if not resp.ok:
            logger.error("Failed to fetch GitHub OAuth response")
            return False

        github_info = resp.json()
        provider = blueprint.name
        provider_id = str(github_info["id"])
        username = github_info["login"]
        name = github_info["name"]
        email = github_info.get("email")

        oauth = blueprint.backend.get_oauth_token_by_provider(
            blueprint.name, provider_id)

        user = None
        user_id = None
        if oauth:
            user = blueprint.backend.get_account(oauth.user_id)
            user_id = oauth.user_id
        else:
            logger.info("Creating user's account")
            user = blueprint.backend.create_account(username, name, email)
            user_id = user.id
            logger.info("Account created {}".format(user_id))

        if not oauth:
            oauth = blueprint.backend.create_oauth_token(
                user_id, provider, provider_id, token)
            logger.info("Created OAuth token {}".format(oauth.id))

        login_user(user, remember=True)

        return False

    @oauth_error.connect_via(blueprint)
    def github_error(blueprint: Blueprint, error, error_description=None,
                     error_uri=None):
        msg = (
            "OAuth error from {name}! "
            "error={error} description={description} uri={uri}"
        ).format(
            name=blueprint.name,
            error=error,
            description=error_description,
            uri=error_uri,
        )
        print(msg)
